;  Class:  CDA 3101. Summer 1989
;  Assn:   mc68000 "trees.asm"
;
;  decription:
;
;  This interactive program takes integer values from the
;  keyboard in the range of 1-9999 and uses them to build a
;  binary search tree.  The user types [RETURN] after each 
;  integer is entered and [0][RETURN] when the data entry
;  is complete.  The program then performs pre- in- and
;  postorder traversals of the tree, sending the output to
;  the screen.
;  
;  registers used...
;
;  d0: gets character from keyboard
;  d1: buffer to hold character input
;  d2: used in char to digit conversion
;  d3: used in char to digit conversion
;  d7: counter holds number of digits entered  
;
;  a6: root pointer of the tree
;  a5: pointer used to traverse the tree
;  a4: used as stack pointer for traversals
;  a0: used to print messages to the screen
;
; initialization
;
prog       equ     $1000  ;program start address
stack      equ     $3000  ;program stack for traversals
data       equ     $0     ;data field in tree nodes
left       equ     $4     ;left pointer in tree nodes
right      equ     $8     ;right pointer in tree nodes
;
init       org     prog
           movea.l #0,a6           ;root pointer=nil
           movea.l #stack,a4       ;a4 ^ program stack
;
start      lea     inptpmpt,a0            ;a0=input prompt address
           movem.l a4-a6/d7/d0-d3,reg_strg ;save registers
           bsr     WriteString             ;write prompt
           movem.l reg_strg,a4-a6/d7/d0-d3 ;restore registers
;
; main takes input from the keyboard and evaluates it, supporting 
; the backspace key to delete mistakes.  No more than four digits
; may be typed as input, and when [return] is pressed, the string
; is converted into a numeric value and is placed in the tree.
; typing [0] and the [return] will cause the program to do pre-
; in- and postorder traversals of the tree.
;
main       move.l  #'0000',d1       ;input char buffer=ascii zeros
           clr.l   d7                   ;character input count=0
get_input  clr.l   d0                   ;empty the input buffer
           movem.l a4-a6/d7/d1,reg_strg ;save registers
           bsr     ReadChar             ;d0=input key
           movem.l reg_strg,a4-a6/d7/d1 ;restore registers
           addi.b  #1,d7                ;inc character count
;
           cmpi.b  #13,d0           ;is character a CR?
           bne     del_key          ;if no, do next check...
ret_key    cmpi.b  #1,d7            ;is CR on an empty line?
           bne     chr2digt         ;if no,handle the num. input
           subi.b  #1,d7            ;CR is on empty line, so dec
           bra     get_input        ;the char count and get input
;
del_key    cmpi.b  #8,d0            ;is character a BS?
           bne     num_keys         ;if no, do next check...
           cmpi.b  #1,d7            ;is BS on a empty line?
           beq     d_k_a               ;if yes, ignore the BS 
           movem.l a4-a6/d7/d1,reg_strg ;save registers
           bsr     WriteChar           ;write the BS
           move.b  #' ',d0             ;move space to d0
           bsr     WriteChar           ;write space to erase #
           move.b  #8,d0               ;move BS to d0
           bsr     WriteChar           ;BS to adjust screen ptr
           movem.l reg_strg,a4-a6/d7/d1 ;restore registers
           move.b  #'0',d1       ;repl.deleted # with 0 in buffer 
           ror.l   #8,d1         ;ror buff.back to proper position
           subi.b  #2,d7         ;decrement charcount twice
           bra     get_input     ;bra to get next character
d_k_a      subi.b  #1,d7         ;BS is on empty line, so dec
           bra     get_input     ;charcount and bra to get next
;
num_keys   cmpi.b  #'0',d0       ;\ if the character entered is
           blt     invalid       ; \  not between '0' and '9'
           cmpi.b  #'9',d0       ; / it is invalid input so
           bgt     invalid       ;/  ignore it...
           cmpi.b  #5,d7         ;have>4 char been entered?
           bne     n_k_a         ;no, so input is good
invalid    subi.b  #1,d7         ;invalid input, so dec charcount
           bra     get_input            ;and bra to get next char
n_k_a      movem.l a4-a6/d7/d1,reg_strg    ;save registers
           bsr     WriteChar            ;write the input to screen
           movem.l reg_strg,a4-a6/d7/d1    ;restore registers
           rol.l   #8,d1         ;ror to make slot for char input
           move.b  d0,d1         ;move the input into the slot
           bra     get_input     ;bra to get next input character
;
; chr2digt converts the ascii characters for an integer into the 
; integer itself by subtracting each ascii digit from ascii '0'
; and then computing the power series expansion 
;
chr2digt   clr.l   d2                   ;clr buffer
           clr.l   d3                   ;clr buffer
           movem.l a4-a6/d1-d3,reg_strg ;save registers
           bsr     WriteEOL             ;write EOL
           movem.l reg_strg,a4-a6/d1-d3 ;restore registers
           cmpi.l  #'0000',d1        ;if input=0 [return]
           beq     traverse          ; then traverse the tree
           subi.b  #'0',d1           ;change lo byte to number
           move.b  d1,d3             ;move it to buffer
           ror.l   #8,d1             ;ror next digit into place
           subi.b  #'0',d1           ;change it to a number
           move.b  d1,d2             ;move it to buffer
           mulu    #10,d2            ;multiply buffer by 10
           add.w   d2,d3             ;and add it to the 1's digit
           ror.l   #8,d1             ;ror next digit into place
           subi.b  #'0',d1           ;convert to a numeric value
           clr.l   d2                ;clear buffer
           move.b  d1,d2             ;move 100's digit to buffer
           mulu    #100,d2           ;multiply by 100
           add.l   d2,d3             ;add to 10's and 1's digit
           ror.l   #8,d1             ;ror next digit into place
           subi.b  #'0',d1           ;convert it to a #
           clr.l   d2                ;clr buffer
           move.b  d1,d2             ;move 1000's digit to buffer
           mulu    #1000,d2          ;mult by 1000 to get value
           add.l   d2,d3      ;add buffers,now d3=# to go in tree
;
; build_tree builds a binary search tree with the integers passed 
; from the chr2digt section, using Malloc to allocate new nodes.
; each node is 12 bytes in length and has three fields:
;                    data field (4 bytes)
;                   left pointer (4 bytes)
;                  right pointer (4 bytes)
;
build_tree move.l  #12,d0            ;want to allocate 12 bytes
           movem.l a4-a6/d3,reg_strg ;save registers
           bsr     malloc            ;allocate the new node
           movem.l reg_strg,a4-a6/d3 ;restore registers
           move.l  #0,left(a0)       ;set leftptr of node=nil
           move.l  #0,right(a0)      ;set rightptr of node=nil
           movea   a6,a5             ;traversal ptr=root
           cmpa.l  #0,a6             ;has the tree been built?
           bne     add_node          ;if yes, then add a new node
;
first_node movea   a0,a6             ;no, so this is the first
           move.l  d3,data(a0)       ;move # to data field
           bra     main              ;bra to get next integer
;
add_node   cmp.l   data(a5),d3       ;is new #<current node's #
           blt     go_left           ;if yes, then move ptr left
go_right   cmpi.l  #0,right(a5)      ;is rightptr of node=nil?
           bne     g_r_a             ;if no,then move ptr right
           move.l  a0,right(a5)      ;rptr=nil,so rptr=rightnode
           move.l  d3,data(a0)       ;newnode gets new data #
           bra     main              ;bra to get next integer
;
g_r_a      movea.l right(a5),a5      ;move pointer right one node
           bra     add_node          ;bra to compare values
;
go_left    cmpi.l  #0,left(a5)       ;is leftptr=nil
           bne     g_l_a             ;if no, then move ptr left
           move.l  a0,left(a5)       ;lptr=nil,so lptr=leftnode
           move.l  d3,data(a0)       ;newnode gets new data #
           bra     main              ;bra to get next integer
;
g_l_a      movea.l left(a5),a5       ;move ptr left one node
           bra     add_node          ;bra to compare values
;
; the traverse section of the program performs the pre-, in-
; and postorder traversals of the binary tree
;
traverse   cmpa.l  #0,a6           ;has tree been built?
           beq     init            ;if no, start over
           lea     pretitle,a0     ;a0^preorder title
           movem.l a4-a6,reg_strg  ;save registers
           bsr     WriteString     ;write title
           movem.l reg_strg,a4-a6  ;restore registers
           movea   a6,a5           ;traversal ptr=rootptr
           movea.l #stack,a4       ;a4^stack
;
; preorder is an iterative routine which prints the nodes of 
; each subtree in the order parent node, left child, right child
;
preorder   move.l  data(a5),d0     ;move node to d0
           bsr     digt2char       ; and print it to screen
           cmpi.l  #0,left(a5)     ;is leftptr=nil?
           bne     pre_left        ;if no, then move left
           cmpi.l  #0,right(a5)    ;is rightptr=nil?
           bne     pre_right       ;if no, then move right
           cmpa.l  #stack,a4       ;is a4 ^ to top of stack?
           beq     end_pre         ;if yes,then preorder finished
           movea.l -(a4),a5        ;pop top element & put in ptr
           bra     preorder        ;bra to continue
;
pre_left   cmpi.l  #0,right(a5)    ;is rightptr=nil
           beq     p_l_a           ;if yes, then move right
           move.l  right(a5),(a4)+ ;no,so push ptr to stack
p_l_a      movea.l left(a5),a5     ;and move right
           bra     preorder        ;bra to continue
pre_right  movea.l right(a5),a5    ;move right one node
           bra     preorder        ;bra to continue
end_pre    bsr     waiting         ;done, so wait for key
;
           movea   a6,a5           ;traversal ptr=rootptr
           movea.l #stack,a4       ;a4 ^ to stack
           lea     intitle,a0      ;a0 ^ inorder title
           movem.l a4-a6,reg_strg  ;save registers
           bsr     WriteString     ;write inorder title
           movem.l reg_strg,a4-a6  ;restore rigisters
           move.l  #0,(a4)+        ;bottom element of stack=0
;
; inorder is an iterative routine which prints each subtree in 
; the order left child, parent node, right child
;
inorder    move.l  a5,(a4)+        ;push current node
           cmpi.l  #0,left(a5)     ;is leftptr=nil?
           bne     in_left         ;if no,then move left
in_pop     movea.l -(a4),a5        ;pop stack into a5
           cmpa.l  #0,a5           ;is this the bottom of stack?
           beq     end_in          ;if yes, then inorder done
           move.l  data(a5),d0     ;d0=data in current node
           bsr     digt2char       ;print data value to screen
           cmpi.l  #0,right(a5)    ;is rightptr=nil?
           bne     in_right        ;if no, then move right
           bra     in_pop          ;pop next element
;
in_left    movea.l left(a5),a5     ;move ptr left one node
           bra     inorder         ;bra to continue traversal
in_right   movea.l right(a5),a5    ;move ptr right one node
           bra     inorder         ;bra to continue traversal
end_in     bsr     waiting         ;inorder done...wait for key
;
           movea   a6,a5           ;traversal ptr=rootptr
           movea.l #stack,a4       ;a4^to the stack
           lea     posttitle,a0    ;a0^to the postorder title
           movem.l a4-a6,reg_strg  ;save registers
           bsr     WriteString     ;write postorder title
           movem.l reg_strg,a4-a6  ;restore registers
;
           bsr     postorder       ;do postorder traversal
           bsr     waiting         ;then wait for keypress
           bra     init            ;and rerun the program
;
; postorder is a recursive routine that prints each subtree in 
; the order left child, right child, parent node
;
postorder  cmpa.l  #0,a5           ;does a5^ to nil?
           beq     pop             ;if yes, then pop stack
           move.l  a5,(a4)+        ;push current node to stack
           movea.l left(a5),a5     ;move left one node
           bsr     postorder       ;call subroutine with leftnode
           move.l  a5,(a4)+        ;push current node to stack
           movea.l right(a5),a5    ;move right one node
           bsr     postorder       ;call subroutine w/ rightnode
           move.l  data(a5),d0     ;d0=current node's data
           bsr     digt2char       ;print data to screen
pop        movea.l -(a4),a5        ;pop element and put it in a5
           rts                     ;return from subroutine
;
; called at the end of each traversal, the waiting routine prints 
; "<press any key to continue>" and waits for a key to be pressed
;
waiting    movem.l a4-a6,reg_strg  ;save registers
           lea     waitmsg,a0      ;a0^to wait message
           bsr     WriteString     ;write wait message
           bsr     ReadChar        ;wait for keypress
           movem.l reg_strg,a4-a6  ;restore registers
           rts                     ;and return
;
; digt2char converts a numeric integer to its ascii equivalent 
; and prints it to the screen one digit at a time
;
digt2char  clr.l   d1              ;clear buffer1
           move.l  d0,d1           ;move integer to buffer
           clr.l   d0              ;clear buffer0
           divu    #1000,d1        ;isolate 1000's place
           cmpi.w  #0,d1           ;\if zero, 
           beq     d2c_a           ;/     do next division
           addi.b  #'0',d1         ;convert # to ascii char
           move.b  d1,d0           ;move character to d0 to print
           movem.l a4-a6/d1,reg_strg  ;save registers
           bsr     WriteChar          ;write char (1000's digit)
           movem.l reg_strg,a4-a6/d1  ;restore registers
           move.w  #0,d1           ;clear lo word of buffer1
d2c_a      swap    d1              ;swap remainder to lo word
           divu    #100,d1         ;isolate 100's digit
           cmpi.w  #0,d1           ;\ if zero,
           beq     d2c_b           ;/   do next division
           addi.b  #'0',d1         ;change # to ascii character
           move.b  d1,d0           ;move character to d0 to print
           movem.l a4-a6/d1,reg_strg   ;save registers
           bsr     WriteChar           ;write char (100's digit)
           movem.l reg_strg,a4-a6/d1   ;restore registers
           move.w  #0,d1           ;clear lo word of buffer1
d2c_b      swap    d1              ;swap remainder to lo word
           divu    #10,d1          ;isolate 10's place digit
           cmpi.w  #0,d1           ;\ if zero,
           beq     d2c_c           ;/   do next division
           addi.b  #'0',d1         ;change # to ascii character
           move.b  d1,d0           ;move char to d0 to print
           movem.l a4-a6/d1,reg_strg    ;save registers
           bsr     WriteChar            ;write char (10's digit)
           movem.l reg_strg,a4-a6/d1    ;restore registers
d2c_c      swap    d1              ;swap remainder (1's)to lo
           addi.b  #'0',d1         ;convert # to ascii character
           move.b  d1,d0           ;move char to d0 to print
           movem.l a4-a6/d1,reg_strg    ;save registers
           bsr     WriteChar            ;write char (1's digit)
           bsr     WriteEOL             ;write CR and LF
           movem.l reg_strg,a4-a6/d1    ;restore registers
           rts                       ;return from subroutine
;
; prompts, messages, and register storage follow...
;
inptpmpt   dc.b    13,10,'trees.asm:',13,10 
           dc.b    'to build the tree, input integers',13,10
           dc.b    'between 1 and 9999, pressing [RETURN]',13,10
           dc.b    'after each one. (0 [RETURN] to end)',13,10,0
intitle    dc.b    13,10,'inorder traversal...',13,10,0
pretitle   dc.b    13,10,'preorder traversal...',13,10,0
posttitle  dc.b    13,10,'postorder traversal...',13,10,0
waitmsg    dc.b    13,10,'<press any key to continue>',13,10,0
reg_strg   ds.l    8
           include "samples.asm"
           end


